#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
CLI for listing, enabling, disabling intake drivers
"""

from intake import __version__
from intake.cli.util import Subcommand
from intake.source.discovery import drivers

import logging
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


class Drivers(Subcommand):
    """
    List, enable, and disable intake drivers.
    """

    name = "drivers"

    def initialize(self):
        sub_parser = self.parser.add_subparsers()

        list = sub_parser.add_parser(
            'list',
            help='Show all intake drivers, whether enabled, disabled, '
                 'or directly inserted into the registry'
        )
        list.add_argument(
            '-v', '--verbose', action='store_true', help='Show module path.')
        list.set_defaults(invoke=self._list)

        enable = sub_parser.add_parser('enable', help='Enable an intake driver.')
        enable.add_argument('name', type=str, help='Driver name')
        enable.add_argument('driver', type=str, default=None, nargs='?',
                            help='Module path and class name, as in '
                                 'package.submodule.ClassName')
        enable.set_defaults(invoke=self._enable)

        disable = sub_parser.add_parser(
            'disable', help='Disable one or more intake drivers.')
        disable.add_argument('names', type=str, help='Driver names', nargs='+')
        disable.set_defaults(invoke=self._disable)

    def invoke(self, args):
        self.parser.print_help()

    def _list(self, args):
        if drivers.do_scan:
            print("Package scan:")
            for k, v in drivers.scanned.items():
                print(f'{k:<30}{v.__module__}.{v.__name__}')
            print()

        print("Entrypoints:")
        eps = [ep for ep in drivers.from_entrypoints()
               if ep.name not in drivers.disabled()]
        if eps:
            for v in eps:
                print(f'{v.name:<30}{v.module_name}:{v.object_name}')
        else:
            print("<none>")
        print()

        print("From Config:")
        eps = [ep for ep in drivers.from_conf()
               if ep.name not in drivers.disabled()]
        if eps:
            for v in eps:
                if v.name not in drivers.disabled():
                    print(f'{v.name:<30}{v.module_name}:{v.object_name}')
        else:
            print("<none>")
        print()

        print("Disabled: ", drivers.disabled() or "<none>")

    def _enable(self, args):
        drivers.enable(args.name, args.driver)

    def _disable(self, args):
        for name in args.names:
            drivers.disable(name)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os

if os.getenv("COVERAGE_PROCESS_START", False):
    import coverage
    coverage.process_startup()



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from packaging.version import Version

gl = globals()


def do_import():
    try:
        import panel as pn
        import hvplot
        error = (Version(pn.__version__) < Version("0.9.5")
                   or Version(hvplot.__version__) < Version("0.8.1"))
    except ImportError:
        error = True

    if error:
        raise RuntimeError("Please install panel and hvplot to use the GUI\n"
                           "`conda install -c conda-forge 'panel>=0.9.5' 'hvplot>=0.8.1'`")

    from .gui import GUI
    css = """
    .scrolling {
      overflow: scroll;
    }
    """
    pn.config.raw_css.append(css)  # add scrolling class from css (panel GH#383, GH#384)
    pn.extension()
    gl['instance'] = GUI()


def __getattr__(attr):
    if attr in {'instance', 'gui'}:
        do_import()
    return gl[attr]


def output_notebook(inline=True, logo=False):
    """
    Load the notebook extension

    Parameters
    ----------
    inline : boolean (optional)
        Whether to inline JS code or load it from a CDN
    logo : boolean (optional)
        Whether to show the logo(s)
    """
    try:
        import hvplot
    except ImportError:
        raise ImportError("The intake plotting API requires hvplot."
                          "hvplot may be installed with:\n\n"
                          "`conda install -c pyviz hvplot` or "
                          "`pip install hvplot`.")
    import holoviews as hv
    return hv.extension('bokeh', inline=inline, logo=logo)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import importlib
import re
import logging
import os
import warnings

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .source import registry
from .catalog.base import Catalog

imports = {
    "DataSource": "intake.source.base:DataSource",
    'Schema': "intake.source.base:Schema",
    "load_combo_catalog": "intake.catalog.default:load_combo_catalog",
    "upload": "intake.container:upload",
    "gui": "intake.interface:instance",
    "interface": "intake.interface",
    "cat": "intake.catalog:builtin",
    "output_notebook": "intake.interface:output_notebook",
    "register_driver": "intake.source:register_driver",
    "unregister_driver": "intake.source:unregister_driver",
}
openers = set()
logger = logging.getLogger('intake')


def __getattr__(attr):
    """Lazy attribute propagator

    Defers inputs of functions until they are needed, according to the
    contents of the ``imports`` (submodules and classes) and ``openers``
    (functions which instantiate data sources directly)
    dicts. All keys in ``openers``
    must start with "open_", else they will be ignored.
    """
    gl = globals()
    if attr in openers and attr[:5] == "open_":
        driver = registry[attr[5:]]  # "open_..."
        gl[attr] = driver
    else:
        if attr in gl:
            return gl[attr]
        elif attr in imports:
            dest = imports[attr]
            modname = dest.split(":", 1)[0]
            logger.debug("Importing: %s" % modname)
            mod = importlib.import_module(modname)
            if ":" in dest:
                gl[attr] = getattr(mod, dest.split(":")[1])
            else:
                gl[attr] = mod
    if attr == "__all__":
        return __dir__()
    try:
        return gl[attr]
    except KeyError:
        raise AttributeError(attr)


def __dir__(*_, **__):
    return sorted(list(globals()) + list(openers) + list(imports))


def make_open_functions():
    """From the current state of ``registry``, create open_* functions"""
    from .source.discovery import drivers

    for name in drivers.enabled_plugins():

        func_name = 'open_' + name
        if not func_name.isidentifier():
            # primitive name normalization
            func_name = re.sub('[-=~^&|@+]', '_', func_name)
        if func_name.isidentifier():
            # stash name for dir() and later fetch
            openers.add(func_name)
        else:
            warnings.warn('Invalid Intake plugin name "%s" found.', name, stacklevel=2)


make_open_functions()


def open_catalog(uri=None, **kwargs):
    """Create a Catalog object

    Can load YAML catalog files, connect to an intake server, or create any
    arbitrary Catalog subclass instance. In the general case, the user should
    supply ``driver=`` with a value from the plugins registry which has a
    container type of catalog. File locations can generally be remote, if
    specifying a URL protocol.

    The default behaviour if not specifying the driver is as follows:

    - if ``uri`` is a single string ending in "yml" or "yaml", open it as a
      catalog file
    - if ``uri`` is a list of strings, a string containing a glob character
      ("*") or a string not ending in "y(a)ml", open as a set of catalog
      files. In the latter case, assume it is a directory.
    - if ``uri`` begins with protocol ``"intake:"``, connect to a remote
      Intake server
    - if ``uri`` is ``None`` or missing, create a base Catalog object without entries.

    Parameters
    ----------
    uri: str or pathlib.Path
        Designator for the location of the catalog.
    kwargs:
        passed to subclass instance, see documentation of the individual
        catalog classes. For example, ``yaml_files_cat`` (when specifying
        multiple uris or a glob string) takes the additional
        parameter ``flatten=True|False``, specifying whether all data sources
        are merged in a single namespace, or each file becomes
        a sub-catalog.

    See also
    --------
    intake.open_yaml_files_cat, intake.open_yaml_file_cat,
    intake.open_intake_remote
    """
    driver = kwargs.pop('driver', None)
    if isinstance(uri, os.PathLike):
        uri = os.fspath(uri)
    if driver is None:
        if uri:
            if ((isinstance(uri, str) and "*" in uri)
                    or ((isinstance(uri, (list, tuple))) and len(uri) > 1)):
                # glob string or list of files/globs
                driver = 'yaml_files_cat'
            elif isinstance(uri, (list, tuple)) and len(uri) == 1:
                uri = uri[0]
                if "*" in uri[0]:
                    # single glob string in a list
                    driver = 'yaml_files_cat'
                else:
                    # single filename in a list
                    driver = 'yaml_file_cat'
            elif isinstance(uri, str):
                # single URL
                if uri.startswith('intake:'):
                    # server
                    driver = 'intake_remote'
                else:
                    if uri.endswith(('.yml', '.yaml')):
                        driver = 'yaml_file_cat'
                    else:
                        uri = uri.rstrip('/') + '/*.y*ml'
                        driver = 'yaml_files_cat'
            else:
                raise ValueError("URI not understood: %s" % uri)
        else:
            # empty cat
            driver = 'catalog'
    if '_file' not in driver:
        kwargs.pop('fs', None)
    if driver not in registry:
        raise ValueError('Unknown catalog driver (%s), supply one of: %s'
                         % (driver, list(sorted(registry))))
    return registry[driver](uri, **kwargs)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import datetime
import itertools
from .base import RemoteSource, Schema, get_partition


class RemoteArray(RemoteSource):
    """nd-array on an Intake server"""
    name = 'remote_ndarray'
    container = 'ndarray'

    def __init__(self, url, headers, **kwargs):
        super(RemoteArray, self).__init__(url, headers, **kwargs)
        self.npartitions = kwargs['npartitions']
        self.shape = tuple(kwargs['shape'])
        self.metadata = kwargs['metadata']
        self.dtype = kwargs['dtype']
        self.chunks = tuple(tuple(c)
                            for c in tuple(kwargs.get('chunks', (-1, ))))
        self.arr = None
        self._schema = Schema(npartitions=self.npartitions,
                              extra_metadata=self.metadata,
                              dtype=self.dtype,
                              shape=self.shape,
                              chunks=self.chunks)

    def _load_metadata(self):
        import dask.array as da
        if self.arr is None:
            name = 'remote-array-' + self._source_id
            nparts = (range(len(n)) for n in self.chunks)
            dask = {
                (name, ) + part: (get_partition, self.url, self.headers,
                                  self._source_id, self.container, part)

                for part in itertools.product(*nparts)
            }
            self.arr = da.Array(dask, name=name, chunks=self.chunks,
                                dtype=self.dtype, shape=self.shape)

        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.arr.blocks[i].compute()

    def read_partition(self, i):
        self._load_metadata()
        return self._get_partition(i)

    def read(self):
        self._load_metadata()
        return self.arr.compute()

    def to_dask(self):
        self._load_metadata()
        return self.arr

    def _close(self):
        self.arr = None

    @staticmethod
    def _persist(source, path, component=None, storage_options=None,
                 **kwargs):
        """Save array to local persistent store

        Makes a parquet dataset out of the data using zarr.
        This then becomes a data entry in the persisted datasets catalog.
        Only works locally for the moment.

        Parameters
        ----------
        source: a DataSource instance to save
        name: str or None
            Key to refer to this persisted dataset by. If not given, will
            attempt to get from the source's name
        kwargs: passed on to zarr array creation, see
        """
        try:
            arr = source.to_dask()
        except NotImplementedError:
            arr = source.read()
        return RemoteArray._data_to_source(arr, path, component=None,
                                           storage_options=None, **kwargs)

    @staticmethod
    def _data_to_source(arr, path, component=None,  storage_options=None,
                        **kwargs):
        from dask.utils import is_arraylike
        from dask.array import to_zarr, from_array
        from ..source.zarr import ZarrArraySource
        if not is_arraylike(arr):
            raise NotImplementedError
        if not hasattr(arr, 'npartitions'):
            arr = from_array(arr, chunks='auto')
        to_zarr(arr, path, component=None,
                storage_options=storage_options, **kwargs)
        source = ZarrArraySource(path, storage_options, component)
        return source



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
'''
'''

import logging
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports

# External imports

# Intake imports
from intake import open_catalog
from intake.cli.util import Subcommand

#-----------------------------------------------------------------------------
# API
#-----------------------------------------------------------------------------

class Precache(Subcommand):
    ''' Populate caching for catalog entries that define caching.
    '''

    name = "precache"

    def initialize(self):
        self.parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')

    def invoke(self, args):
        catalog = open_catalog(args.uri)
        for entry in list(catalog):
            try:
                s = catalog[entry]
                s.read()
                if s.cache:
                    print("Caching for entry %s" % entry)
            except Exception as e:
                print("Skipping {} due to {}".format(entry, e))
